// ----------------------------------------------------------------------------
//
//  Copyright (C) 2020 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#include <string.h>
#include <math.h>
#include "dominance1.h"


#define DEG2RAD 0.01745329f
#define PIFLOAT 3.141592f


const LADSPA_PortDescriptor Ladspa_Dominance1::portdescr [NPORT] = 
{
    LADSPA_PORT_INPUT  | LADSPA_PORT_AUDIO,
    LADSPA_PORT_INPUT  | LADSPA_PORT_AUDIO,
    LADSPA_PORT_INPUT  | LADSPA_PORT_AUDIO,
    LADSPA_PORT_INPUT  | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_INPUT  | LADSPA_PORT_CONTROL,
    LADSPA_PORT_INPUT  | LADSPA_PORT_CONTROL
};


const char * const Ladspa_Dominance1::portnames [NPORT] = 
{
    "in.0", "in.1", "in.2", "in.3",
    "out.0", "out.1", "out.2", "out.3",
    "Direction", "Dominance" 
};


const LADSPA_PortRangeHint Ladspa_Dominance1::porthints [NPORT] = 
{
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0, -180, 180 },
    { LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0, -0.75, 0.75 }
};


LADSPA_Handle Ladspa_Dominance1::create (const struct _LADSPA_Descriptor *desc, unsigned long rate)
{
    return new Ladspa_Dominance1 (rate);
}


Ladspa_Dominance1::Ladspa_Dominance1 (unsigned long fsam) :
    LadspaPlugin (fsam)
{
    int i, j;

    for (i = 0; i < 4; i++)
    {
	for (j = i; j < 4; j++)
	{
	    _c [i][j] = (i == j) ? 1.0f : 0.0f;
	}
    }
}


void Ladspa_Dominance1::setport (unsigned long port, LADSPA_Data *data)
{
    _port [port] = data;
}


void Ladspa_Dominance1::active (bool act)
{
}


void Ladspa_Dominance1::runproc (unsigned long len, bool add)
{
    int    k;
    float  cd, sd, a, b;
    float  w, y, z, x;
    float  c [4][4], d [4][4];
     
    a = _port [CTL_DIR][0] * PIFLOAT / 180;
    cd = cosf (a);
    sd = sinf (a);
    a = _port [CTL_DOM][0];
    b = sqrtf (1 - a * a);

    c [0][1] = -a * sd;
    c [0][3] =  a * cd;
    c [1][1] = b * cd * cd + sd * sd;
    c [1][3] = cd * sd * (b - 1);
    c [2][2] = b;
    c [3][3] = b * sd * sd + cd * cd;

    d [0][1] = (c [0][1] - _c [0][1]) / len;
    d [0][3] = (c [0][3] - _c [0][3]) / len;
    d [1][1] = (c [1][1] - _c [1][1]) / len;
    d [1][3] = (c [1][3] - _c [1][3]) / len;
    d [2][2] = (c [2][2] - _c [2][2]) / len;
    d [3][3] = (c [3][3] - _c [3][3]) / len;

    _c [0][1] = c [0][1]; 
    _c [0][3] = c [0][3]; 
    _c [1][1] = c [1][1]; 
    _c [1][3] = c [1][3]; 
    _c [2][2] = c [2][2]; 
    _c [3][3] = c [3][3]; 
    
    for (k = 0; k < (int) len; k++)
    {
	c [0][1] += d [0][1];
	c [0][3] += d [0][3];
	c [1][1] += d [1][1];
	c [1][3] += d [1][3];
	c [2][2] += d [2][2];
	c [3][3] += d [3][3];

	w = _port [INPW][k]; 
	y = _port [INPY][k]; 
	z = _port [INPZ][k]; 
	x = _port [INPX][k]; 

	_port [OUTW][k] =            w + c [0][1] * y + c [0][3] * x;
	_port [OUTY][k] = c [0][1] * w + c [1][1] * y + c [1][3] * x;
	_port [OUTZ][k] = c [2][2] * z;
	_port [OUTX][k] = c [0][3] * w + c [1][3] * y + c [3][3] * x;
    }
}
